# -*- coding: utf-8 -*-

import os
import logging
from typing import Callable

import pytest
from pydantic import Field

from onion_config import ConfigLoader, BaseConfig


logger = logging.getLogger(__name__)


@pytest.fixture
def config_loader():
    _config_loader = ConfigLoader()

    yield _config_loader

    del _config_loader


@pytest.fixture
def configs_dir(tmp_path):
    _tmp_configs_dir_pl = tmp_path / "configs"
    _tmp_configs_dir_pl.mkdir()
    _tmp_json_file_pl = (_tmp_configs_dir_pl / "test.json").resolve()
    _tmp_json_file_pl.write_text(
        '{"json_test": {"str_val": "some_value", "int_val": 123} }'
    )
    _tmp_yaml_file_pl = (_tmp_configs_dir_pl / "test.yml").resolve()
    _tmp_yaml_file_pl.write_text("yaml_test: true")
    _tmp_configs_dir = str(_tmp_configs_dir_pl)
    _expected = {
        "json_test": {"str_val": "some_value", "int_val": 123},
        "yaml_test": True,
    }

    yield _tmp_configs_dir, _expected

    del (
        _tmp_configs_dir_pl,
        _tmp_json_file_pl,
        _tmp_yaml_file_pl,
        _tmp_configs_dir,
        _expected,
    )


def test_init(config_loader):
    logger.info("Testing initialization of 'ConfigLoader'...")

    assert isinstance(config_loader, ConfigLoader)
    assert isinstance(config_loader.configs_dir, str)
    assert config_loader.configs_dir == ConfigLoader._CONFIGS_DIR
    assert issubclass(config_loader.config_schema, BaseConfig)
    assert config_loader.config_schema == BaseConfig
    assert isinstance(config_loader.required_envs, list)
    assert config_loader.required_envs == []
    assert isinstance(config_loader.pre_load_hook, Callable)
    assert config_loader.pre_load_hook == ConfigLoader._PRE_LOAD_HOOK
    assert isinstance(config_loader.env_file_path, str)
    assert config_loader.env_file_path == ConfigLoader._ENV_FILE_PATH
    assert config_loader.extra_configs_dir == None
    assert isinstance(config_loader.config_data, dict)
    assert config_loader.config_data == {}
    assert config_loader.config == None

    logger.info("Done: Initialization of 'ConfigLoader'.")


def test_load(config_loader):
    logger.info("Testing 'load' method...")

    _config: BaseConfig = config_loader.load()

    assert isinstance(_config, BaseConfig)
    assert config_loader.config == _config

    logger.info("Done: 'load' method.")


@pytest.mark.parametrize(
    "configs_dir, config_schema, pre_load_hook, env_file_path, required_envs, config_data, extra_configs_dir,  expected",
    [
        (
            ConfigLoader._CONFIGS_DIR,
            BaseConfig,
            lambda config_data: config_data,
            ConfigLoader._ENV_FILE_PATH,
            [],
            {},
            None,
            {},
        )
    ],
)
def test_config_load(
    configs_dir,
    config_schema,
    pre_load_hook,
    env_file_path,
    required_envs,
    config_data,
    extra_configs_dir,
    expected,
):
    logger.info("Testing main config cases...")

    _config = ConfigLoader(
        configs_dir=configs_dir,
        config_schema=config_schema,
        pre_load_hook=pre_load_hook,
        env_file_path=env_file_path,
        required_envs=required_envs,
        config_data=config_data,
        extra_configs_dir=extra_configs_dir,
    ).load()

    assert isinstance(_config, config_schema)
    assert _config.model_dump() == expected

    logger.info("Done: Main config cases.")


@pytest.mark.parametrize(
    "content, expected",
    [
        ("ENV=test", "test"),
        ("DEBUG=false", "false"),
        ("TEST_ENV_VAR=123", "123"),
    ],
)
def test_load_dotenv(tmp_path, config_loader, content, expected):
    logger.info("Testing 'load_dotenv' method...")

    _tmp_envs_dir_pl = tmp_path / "envs"
    _tmp_envs_dir_pl.mkdir()
    _tmp_env_file_pl = (_tmp_envs_dir_pl / ".env").resolve()
    _tmp_env_file_pl.write_text(content)
    _tmp_env_path = str(_tmp_env_file_pl)

    config_loader.env_file_path = _tmp_env_path
    config_loader._load_dotenv()
    _env_var = content.split("=")[0]

    assert config_loader.env_file_path == _tmp_env_path
    assert os.getenv(_env_var) == expected

    logger.info("Done: 'load_dotenv' method.")


def test_check_required_envs(config_loader):
    logger.info("Testing 'check_required_envs' method...")

    os.environ["REQUIRED_ENV_VAR"] = "required_value"

    config_loader.required_envs = ["REQUIRED_ENV_VAR"]
    config_loader._check_required_envs()

    assert config_loader.required_envs == ["REQUIRED_ENV_VAR"]
    assert os.getenv("REQUIRED_ENV_VAR") == "required_value"

    with pytest.raises(KeyError):
        _none_existent_env_var = "NON_EXISTENT_ENV_VAR"
        config_loader.required_envs = [_none_existent_env_var]
        config_loader._check_required_envs()
        assert os.getenv(_none_existent_env_var) == None

    logger.info("Done: 'check_required_envs' method.")


def test_load_config_files(config_loader, configs_dir):
    logger.info("Testing 'load_config_files' method...")

    _configs_dir, _expected = configs_dir
    config_loader.configs_dir = _configs_dir
    config_loader._load_config_files()

    assert isinstance(config_loader.config_data, dict)
    assert config_loader.config_data["json_test"]["str_val"] == "some_value"
    assert config_loader.config_data["json_test"]["int_val"] == 123
    assert config_loader.config_data["yaml_test"] == True
    assert config_loader.config_data == _expected

    logger.info("Done: 'load_config_files' method.")


def test_load_extra_config_files(tmp_path, config_loader, configs_dir):
    logger.info("Testing 'load_extra_config_files' method...")

    _configs_dir, _expected = configs_dir
    config_loader._load_config_files(configs_dir=_configs_dir)

    _tmp_extra_configs_dir_pl = tmp_path / "extra_configs"
    _tmp_extra_configs_dir_pl.mkdir()
    _tmp_yaml_file_pl = (_tmp_extra_configs_dir_pl / "test.yaml").resolve()
    _tmp_yaml_file_pl.write_text(
        'json_test:\n  str_val: "updated_val"\n  int_val: 321\nextra_test: "extra_value"'
    )
    _tmp_extra_configs_dir = str(_tmp_extra_configs_dir_pl)
    _expected["json_test"]["int_val"] = 321
    _expected["json_test"]["str_val"] = "updated_val"
    _expected["extra_test"] = "extra_value"

    config_loader.extra_configs_dir = _tmp_extra_configs_dir
    config_loader._load_extra_config_files()

    assert isinstance(config_loader.config_data, dict)
    assert config_loader.config_data["json_test"]["str_val"] == "updated_val"
    assert config_loader.config_data["json_test"]["int_val"] == 321
    assert config_loader.config_data["extra_test"] == "extra_value"
    assert config_loader.config_data == _expected

    logger.info("Done: 'load_extra_config_files' method.")


def test_config(config_loader):
    logger.info("Testing 'config' property...")

    class _ConfigSchema(BaseConfig):
        test_var: str = Field("default_val", min_length=2, max_length=32)

    config_loader.config_schema = _ConfigSchema
    config_loader.load()
    print(config_loader.config.model_dump())

    assert isinstance(config_loader.config, _ConfigSchema)
    assert config_loader.config.test_var == "default_val"
    assert config_loader.config.model_dump() == {"test_var": "default_val"}

    config_loader.config = _ConfigSchema(test_var="new_val")
    assert config_loader.config.test_var == "new_val"
    assert config_loader.config.model_dump() == {"test_var": "new_val"}

    with pytest.raises(TypeError):
        config_loader.config = {"test_var": "invalid_val"}
        config_loader.config = "invalid_val"
        config_loader.config = 3.14
        config_loader.config = False
        config_loader.config = None

    logger.info("Done: 'config' property.")


def test_config_schema(config_loader):
    logger.info("Testing 'config_schema' property...")

    class _ConfigSchema(BaseConfig):
        pass

    config_loader.config_schema = _ConfigSchema
    assert config_loader.config_schema == _ConfigSchema

    with pytest.raises(TypeError):
        config_loader.config_schema = "invalid_val"
        config_loader.config_schema = 3.14
        config_loader.config_schema = False
        config_loader.config_schema = None

    logger.info("Done: 'config_schema' property.")


def test_config_data(config_loader):
    logger.info("Testing 'config_data' property...")

    _config_data = {"test_var": "test_val"}
    config_loader.config_data = _config_data

    assert isinstance(config_loader.config_data, dict)
    assert config_loader.config_data == _config_data
    assert config_loader.config_data["test_var"] == "test_val"
    assert config_loader.config_data == _config_data

    with pytest.raises(TypeError):
        config_loader.config_data = _config_data
        config_loader.config_data = "invalid_val"
        config_loader.config_data = 3.14
        config_loader.config_data = False
        config_loader.config_data = None

    logger.info("Done: 'config_data' property.")


def test_configs_dir(config_loader):
    logger.info("Testing 'configs_dir' property...")

    config_loader.configs_dir = "/tmp/pytest/configs_dir"
    assert config_loader.configs_dir == "/tmp/pytest/configs_dir"

    with pytest.raises(TypeError):
        config_loader.configs_dir = 3.14
        config_loader.configs_dir = False
        config_loader.configs_dir = None

    with pytest.raises(ValueError):
        config_loader.configs_dir = ""

    logger.info("Done: 'configs_dir' property.")


def test_extra_configs_dir(config_loader):
    logger.info("Testing 'extra_configs_dir' property...")

    config_loader.extra_configs_dir = "/tmp/pytest/extra_configs_dir"
    assert config_loader.extra_configs_dir == "/tmp/pytest/extra_configs_dir"

    with pytest.raises(TypeError):
        config_loader.extra_configs_dir = 3.14
        config_loader.extra_configs_dir = False
        config_loader.extra_configs_dir = None

    with pytest.raises(ValueError):
        config_loader.extra_configs_dir = ""

    logger.info("Done: 'extra_configs_dir' property.")


def test_env_file_path(config_loader):
    logger.info("Testing 'env_file_path' property...")

    config_loader.env_file_path = "/tmp/pytest/.env"
    assert config_loader.env_file_path == "/tmp/pytest/.env"

    with pytest.raises(TypeError):
        config_loader.env_file_path = 3.14
        config_loader.env_file_path = False
        config_loader.env_file_path = None

    with pytest.raises(ValueError):
        config_loader.env_file_path = ""

    logger.info("Done: 'env_file_path' property.")


def test_required_envs(config_loader):
    logger.info("Testing 'required_envs' property...")

    config_loader.required_envs = ["TEST_ENV_VAR"]
    assert config_loader.required_envs == ["TEST_ENV_VAR"]

    config_loader.required_envs = []
    assert config_loader.required_envs == []

    with pytest.raises(TypeError):
        config_loader.required_envs = "invalid_val"
        config_loader.required_envs = 3.14
        config_loader.required_envs = False
        config_loader.required_envs = None

    with pytest.raises(ValueError):
        config_loader.required_envs = ["TEST_ENV_VAR", 1, None]

    logger.info("Done: 'required_envs' property.")


def test_pre_load_hook(config_loader):
    logger.info("Testing 'pre_load_hook' property...")

    def _pre_load_hook(config_data):
        return config_data

    config_loader.pre_load_hook = _pre_load_hook
    _config_data: dict = config_loader.pre_load_hook(config_loader.config_data)

    assert isinstance(_config_data, dict)
    assert config_loader.pre_load_hook == _pre_load_hook

    with pytest.raises(TypeError):
        config_loader.pre_load_hook = "invalid_val"
        config_loader.pre_load_hook = 3.14
        config_loader.pre_load_hook = False
        config_loader.pre_load_hook = None

    logger.info("Done: 'pre_load_hook' property.")



# -*- coding: utf-8 -*-

from typing import Tuple, Type

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    PydanticBaseSettingsSource,
)


class BaseConfig(BaseSettings):
    model_config = SettingsConfigDict(
        extra="allow", frozen=True, arbitrary_types_allowed=True
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return dotenv_settings, env_settings, init_settings, file_secret_settings



# -*- coding: utf-8 -*-

## Standard libraries
import os
import glob
import json
import logging
from typing import Union, List, Callable, Type

## Third-party libraries
import yaml
from dotenv import load_dotenv
from pydantic import BaseModel, validate_call
from pydantic_settings import BaseSettings

## Internal modules
from ._schema import BaseConfig
from ._utils import deep_merge
from .__version__ import __version__


logger = logging.getLogger(__name__)


class ConfigLoader:
    """A core class of `onion_config` module to use as the main config loader.

    Attributes:
        _ENV_FILE_PATH    (str     ): Default '.env' file path to load. Defaults to '${PWD}/.env'.
        _CONFIGS_DIR      (str     ): Default configs directory. Defaults to '${PWD}/configs'.
        _PRE_LOAD_HOOK    (function): Default lambda function for `pre_load_hook`. Defaults to `lambda config_data: config_data`.

        config            (Union[BaseConfig, BaseSettings, BaseModel]): Main config object (config_schema) for project. Defaults to None.
        config_schema     (Union[Type[BaseConfig], Type[BaseSettings], Type[BaseModel]]): Main config schema class to load and validate configs. Defaults to `BaseConfig`.
        config_data       (dict    ): Loaded data from config files as a dictionary. Defaults to {}.
        configs_dir       (str     ): Main configs directory to load all config files. Defaults to `ConfigLoader._CONFIGS_DIR`.
        extra_configs_dir (str     ): Extra configs directory to load extra config files. Defaults to None, but will use the 'ONION_CONFIG_EXTRA_DIR' environment variable if set.
        env_file_path     (str     ): '.env' file path to load. Defaults to `ConfigLoader._ENV_FILE_PATH`.
        required_envs     (str     ): Required environment variables to check. Defaults to [].
        pre_load_hook     (function): Custom pre-load method, this method will executed before validating `config`. Defaults to `ConfigLoader._PRE_LOAD_HOOK`.

    Methods:
        load()                    : Load and validate every configs into `config`.
        _load_dotenv()            : Loading environment variables from '.env' file, if it's exits.
        _check_required_envs()    : Check required environment variables are exist or not.
        _load_config_files()      : Load config files from `configs_dir` into `config_data`.
        _load_extra_config_files(): Load extra config files from `extra_configs_dir` into `config_data`.
    """

    _ENV_FILE_PATH = os.path.join(os.getcwd(), ".env")
    _CONFIGS_DIR = os.path.join(os.getcwd(), "configs")
    _PRE_LOAD_HOOK = lambda config_data: config_data

    @validate_call
    def __init__(
        self,
        configs_dir: str = _CONFIGS_DIR,
        config_schema: Union[
            Type[BaseConfig], Type[BaseSettings], Type[BaseModel]
        ] = BaseConfig,
        pre_load_hook: Callable = _PRE_LOAD_HOOK,
        env_file_path: str = _ENV_FILE_PATH,
        required_envs: List[str] = [],
        config_data: dict = {},
        extra_configs_dir: Union[str, None] = None,
        auto_load: bool = False,
    ):
        """ConfigLoader constructor method.

        Args:
            configs_dir       (str,      optional): Main configs directory to load all config files. Defaults to `ConfigLoader._CONFIGS_DIR`.
            config_schema     (Union[Type[BaseConfig], Type[BaseSettings], Type[BaseModel]], optional): Main config schema class to load and validate configs. Defaults to `BaseConfig`.
            pre_load_hook     (function, optional): Custom pre-load method, this method will executed before validating `config`. Defaults to `ConfigLoader._PRE_LOAD_HOOK`.
            env_file_path     (str,      optional): '.env' file path to load. Defaults to `ConfigLoader._ENV_FILE_PATH`.
            required_envs     (list,     optional): Required environment variables to check. Defaults to [].
            config_data       (dict,     optional): Custom config data to load before validating `config`. Defaults to {}.
            extra_configs_dir (str,      optional): Extra configs directory to load extra config files. Defaults to None.
            auto_load         (bool,     optional): Auto load configs on init or not. Defaults to False.
        """

        self.configs_dir = configs_dir
        self.config_schema = config_schema
        self.pre_load_hook = pre_load_hook
        self.env_file_path = env_file_path
        self.required_envs = required_envs
        self.config_data = config_data
        if extra_configs_dir:
            self.extra_configs_dir = extra_configs_dir

        if auto_load:
            self.load()

    def load(self) -> Union[BaseConfig, BaseSettings, BaseModel]:
        """Load and validate every configs into `config`.
        Load order:
            1. Load environment variables from '.env' file.
            2. Check required environment variables are exist or not.
            3. Load config files from `configs_dir` into `config_data`.
            4. Load extra config files from `extra_configs_dir` into `config_data`.
            5. Execute `pre_load_hook` method to modify `config_data`.
            6. Init `config_schema` with `config_data` into final `config`.

        Returns:
            Union[BaseConfig, BaseSettings, BaseModel]: Main config object (config_schema) for project.

        Raises:
            KeyError : If a required environment variable does not exist.
            Exception: If `pre_load_hook` or `config_schema` failed to execute.
        """

        self._load_dotenv()
        self._check_required_envs()
        self._load_config_files()
        self._load_extra_config_files()

        try:
            # 5. Execute `pre_load_hook` method:
            self.config_data: dict = self.pre_load_hook(self.config_data)
        except Exception:
            logger.critical("Failed to execute `pre_load_hook` method:")
            raise

        try:
            # 6. Init `config_schema` with `config_data`:
            self.config: Union[
                BaseConfig, BaseSettings, BaseModel
            ] = self.config_schema(**self.config_data)
        except Exception:
            logger.critical("Failed to init `config_schema`:")
            raise

        return self.config

    def _load_dotenv(self):
        """1. Loading environment variables from '.env' file, if it's exits."""

        if os.path.isfile(self.env_file_path):
            load_dotenv(dotenv_path=self.env_file_path, override=True, encoding="utf8")

    def _check_required_envs(self):
        """2. Check if required environment variables exist or not.

        If a required environment variable does not exist, an error is logged and raise an exception.

        Raises:
            KeyError: If a required environment variable does not exist.
        """

        for _env in self.required_envs:
            try:
                os.environ[_env]
            except KeyError:
                logger.critical(f"Missing required '{_env}' environment variable.")
                raise

    @validate_call
    def _load_config_files(self, configs_dir: Union[str, None] = None):
        """3. Load config files from `configs_dir` into `config_data`.

        Args:
            configs_dir (str, optional): Main configs directory to load all config files. Defaults to `ConfigLoader._CONFIGS_DIR`.

        Raises:
            Exception: If failed to load any config file.
        """

        _configs_dir = self.configs_dir
        if configs_dir:
            _configs_dir = configs_dir

        if os.path.isdir(_configs_dir):
            ## Loading all JSON config files from 'configs' directory:
            _json_file_paths = sorted(glob.glob(os.path.join(_configs_dir, "*.json")))
            for _json_file_path in _json_file_paths:
                try:
                    with open(_json_file_path, "r", encoding="utf8") as _json_file:
                        _new_config_dict = json.load(_json_file)
                        self.config_data = deep_merge(
                            self.config_data, _new_config_dict
                        )
                except Exception:
                    logger.critical(
                        f"Failed to load '{_json_file_path}' json config file:"
                    )
                    raise

            ## Loading all YAML config files from 'configs' directory:
            _yaml_file_paths = glob.glob(os.path.join(_configs_dir, "*.yml"))
            _yaml_file_paths.extend(glob.glob(os.path.join(_configs_dir, "*.yaml")))
            _yaml_file_paths.sort()
            for _yaml_file_path in _yaml_file_paths:
                try:
                    with open(_yaml_file_path, "r", encoding="utf8") as _yaml_file:
                        _new_config_dict = yaml.safe_load(_yaml_file)
                        self.config_data = deep_merge(
                            self.config_data, _new_config_dict
                        )
                except Exception:
                    logger.critical(
                        f"Failed to load '{_yaml_file_path}' yaml config file:"
                    )
                    raise

    def _load_extra_config_files(self):
        """4. Load extra config files from `extra_configs_dir` into `config_data`.

        Raises:
            Exception: If failed to load any extra config file.
        """

        if self.extra_configs_dir is None:
            _env_extra_configs_dir = os.getenv("ONION_CONFIG_EXTRA_DIR")
            if _env_extra_configs_dir:
                self.extra_configs_dir = _env_extra_configs_dir

        if self.extra_configs_dir:
            self._load_config_files(configs_dir=self.extra_configs_dir)

    ### ATTRIBUTES ###

    ## config ##
    @property
    def config(self) -> Union[BaseConfig, BaseSettings, BaseModel, None]:
        try:
            return self.__config
        except AttributeError:
            return None

    @config.setter
    def config(self, config: Union[BaseConfig, BaseSettings, BaseModel]):
        if (
            (not isinstance(config, BaseConfig))
            and (not isinstance(config, BaseSettings))
            and (not isinstance(config, BaseModel))
        ):
            raise TypeError(
                f"`config` attribute type {type(config)} is invalid, must be a <class 'BaseConfig'> or `pydantic` <class 'BaseSettings'> or <class 'BaseModel'>."
            )

        self.__config = config

    ## config ##

    ## config_schema ##
    @property
    def config_schema(
        self,
    ) -> Union[Type[BaseConfig], Type[BaseSettings], Type[BaseModel]]:
        try:
            return self.__config_schema
        except AttributeError:
            self.__config_schema = BaseConfig

        return self.__config_schema

    @config_schema.setter
    def config_schema(
        self,
        config_schema: Union[Type[BaseConfig], Type[BaseSettings], Type[BaseModel]],
    ):
        # Check if config_schema is a class (i.e., an instance of type)
        if not isinstance(config_schema, type):
            raise TypeError("`config_schema` must be a class, not an instance.")

        # Check if config_schema is a subclass of BaseConfig, BaseSettings or BaseModel
        if (
            (not issubclass(config_schema, BaseConfig))
            and (not issubclass(config_schema, BaseSettings))
            and (not issubclass(config_schema, BaseModel))
        ):
            _base_class = ""
            if hasattr(config_schema, "__base__"):
                _base_class = config_schema.__base__
            else:
                _base_class = type(config_schema)

            raise TypeError(
                f"`config_schema` attribute base class {_base_class} is invalid, must be a Type[<class 'BaseConfig'>] or `pydantic` Type[<class 'BaseSettings'>, <class 'BaseModel'>]."
            )

        self.__config_schema = config_schema

    ## config_schema ##

    ## config_data ##
    @property
    def config_data(self) -> dict:
        try:
            return self.__config_data
        except AttributeError:
            self.__config_data = {}

        return self.__config_data

    @config_data.setter
    def config_data(self, config_data: dict):
        if not isinstance(config_data, dict):
            raise TypeError(
                f"`config_data` attribute type {type(config_data)} is invalid, must be a <dict>."
            )

        self.__config_data = config_data

    ## config_data ##

    ## configs_dir ##
    @property
    def configs_dir(self) -> str:
        try:
            return self.__configs_dir
        except AttributeError:
            self.__configs_dir = ConfigLoader._CONFIGS_DIR

        return self.__configs_dir

    @configs_dir.setter
    def configs_dir(self, configs_dir: str):
        if not isinstance(configs_dir, str):
            raise TypeError(
                f"`configs_dir` attribute type {type(configs_dir)} is invalid, must be a <str>!"
            )

        configs_dir = configs_dir.strip()
        if configs_dir == "":
            raise ValueError("The `configs_dir` attribute value is empty!")

        self.__configs_dir = configs_dir

    ## configs_dir ##

    ## extra_configs_dir ##
    @property
    def extra_configs_dir(self) -> Union[str, None]:
        try:
            return self.__extra_configs_dir
        except AttributeError:
            return None

    @extra_configs_dir.setter
    def extra_configs_dir(self, extra_configs_dir: str):
        if not isinstance(extra_configs_dir, str):
            raise TypeError(
                f"`extra_configs_dir` attribute type {type(extra_configs_dir)} is invalid, must be a <str>!"
            )

        extra_configs_dir = extra_configs_dir.strip()
        if extra_configs_dir == "":
            raise ValueError("The `extra_configs_dir` attribute value is empty!")

        self.__extra_configs_dir = extra_configs_dir

    ## extra_configs_dir ##

    ## env_file_path ##
    @property
    def env_file_path(self) -> str:
        try:
            return self.__env_file_path
        except AttributeError:
            self.__env_file_path = ConfigLoader._ENV_FILE_PATH

        return self.__env_file_path

    @env_file_path.setter
    def env_file_path(self, env_file_path: str):
        if not isinstance(env_file_path, str):
            raise TypeError(
                f"The 'env_file_path' attribute type {type(env_file_path)} is invalid, must be a <str>!"
            )

        env_file_path = env_file_path.strip()
        if env_file_path == "":
            raise ValueError("The 'env_file_path' attribute value is empty!")

        self.__env_file_path = env_file_path

    ## env_file_path ##

    ## required_envs ##
    @property
    def required_envs(self) -> List[str]:
        try:
            return self.__required_envs
        except AttributeError:
            self.__required_envs = []

        return self.__required_envs

    @required_envs.setter
    def required_envs(self, required_envs: List[str]):
        if not isinstance(required_envs, list):
            raise TypeError(
                f"The 'required_envs' attribute type {type(required_envs)} is invalid, must be a <list>!"
            )

        if not all(isinstance(_val, str) for _val in required_envs):
            raise ValueError(
                f"The 'required_envs' attribute value {required_envs} is invalid, must be a list of <str>!"
            )

        self.__required_envs = required_envs

    ## required_envs ##

    ## pre_load_hook ##
    @property
    def pre_load_hook(self) -> Callable:
        try:
            return self.__pre_load
        except AttributeError:
            self.__pre_load = ConfigLoader._PRE_LOAD_HOOK

        return self.__pre_load

    @pre_load_hook.setter
    def pre_load_hook(self, pre_load_hook: Callable):
        if not callable(pre_load_hook):
            raise TypeError(
                f"`pre_load_hook` argument type {type(pre_load_hook)} is invalid, should be callable <function>!"
            )

        self.__pre_load = pre_load_hook

    ## pre_load_hook ##
    ### ATTRIBUTES ###



# -*- coding: utf-8 -*-

import copy

from pydantic import validate_call


@validate_call
def deep_merge(dict1: dict, dict2: dict) -> dict:
    """Return a new dictionary that's the result of a deep merge of two dictionaries.
    If there are conflicts, values from `dict2` will overwrite those in `dict1`.

    Args:
        dict1 (dict, required): The base dictionary that will be merged.
        dict2 (dict, required): The dictionary to merge into `dict1`.

    Returns:
        dict: The merged dictionary.
    """

    _merged = copy.deepcopy(dict1)
    for _key, _val in dict2.items():
        if (
            _key in _merged
            and isinstance(_merged[_key], dict)
            and isinstance(_val, dict)
        ):
            _merged[_key] = deep_merge(_merged[_key], _val)
        else:
            _merged[_key] = copy.deepcopy(_val)

    return _merged



# -*- coding: utf-8 -*-

from setuptools import setup
from distutils.util import convert_path


_package_name = "onion_config"

_namespace_dict = {}
_version_path = convert_path(f"{_package_name}/__version__.py")
with open(_version_path) as _version_file:
    exec(_version_file.read(), _namespace_dict)
_package_version = _namespace_dict["__version__"]

setup(
    name=_package_name,
    packages=[_package_name],
    version=f"{_package_version}",
    license="MIT",
    description="onion_config is a Python package that allows for easy configuration management. It allows for loading and validating configuration data from environment variables and config files in JSON and YAML formats.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="Batkhuu Byambajav",
    author_email="batkhuu10@gmail.com",
    url="https://github.com/bybatkhuu/mod.python-config",
    download_url=f"https://github.com/bybatkhuu/mod.python-config/archive/v{_package_version}.tar.gz",
    keywords=[
        _package_name,
        "config",
        "configs",
        "dotenv",
        "python-dotenv",
        "pydantic",
        "pydantic-config",
        "pydantic-settings",
        "custom-config",
    ],
    python_requires=">=3.7",
    install_requires=[
        "python-dotenv>=1.0.0,<2.0.0",
        "PyYAML>=6.0.1,<7.0",
        "pydantic[email]>=2.1.1,<3.0.0",
        "pydantic-settings>=2.0.3,<3.0.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)
